verbose = FALSE

setClass("SOAPClientInterface",
           representation(functions = "list",
                          classes = "list"))



capitalizeFirstLetter =
function(word)
{
  els = strsplit(word, "")[[1]]
  els[1] = toupper(els[1])
  paste(els, collapse="")
}

convertToSName =
function(name, useDash = ifelse("UseDashInSOAPNames" %in% names(options()), getOption("UseDashInSOAPNames"), TRUE))
{
  if(useDash)
    return(name)
  
  els = strsplit(name, "_")[[1]]
  if(length(els) == 1)
    return(els)
  
  els[2:length(els)] = sapply(els[-1], capitalizeFirstLetter)

  paste(els, collapse="")
}

dynamicSOAPServer =
function(iface, obj = new("DynamicSOAPServer"))
{
  server = environment(iface@functions[[1]])$".defaultServer"
  obj@server = server
  obj@functions = iface@functions
  
  obj
}


setMethod("$", c("DynamicSOAPServer", "character"),
           function(x, name) {
               f = x@functions[[name]]
                # Need to put this server object as the default server.
                # formals(f)
               f
           })

#setGeneric("names", function(x) standardGeneric("names"))

setMethod("names", "DynamicSOAPServer",
  function(x)
    names(x@functions)
)


defineClasses =
  #
  # namespaceDefs is used in mapSOAPTypeToS as the third argument.
  #
function(types,  where = globalenv(), namespaceDefs = list(), verbose = FALSE)
{
  pending = new.env(hash = TRUE, emptyenv())
  classes = new.env(hash = TRUE, emptyenv())


  defClass =
      function(i) {

         if(is(i, "XMLNode"))
           return()
        
         if(verbose)
            cat("Attempting to define class for", i$name,"\n")

         name = if(is(i, "SOAPType")) i@name else i$name
         
           # if it is already defined, skip this.
         if(!is.null(getClassDef( name, where )))
           return(FALSE)

           # if it is currently pending, then don't do it as we will get recursive calls.
         if(exists(name, pending))
           return(NA)

           # register the type as pending since we are about to define it.
         assign(name, "TRUE", pending)
         on.exit(remove(list = name, envir = pending))

         o = i
         if(!is(i, "SOAPType"))
           i = i$definition
         
         if(is(i, "EnumValuesDef")) {

           elName = paste(name, "Values", sep = "_")
           assign(elName, as.character(i@values), envir = where)
           #XXX No ZZ and no validateEnum!
#           f = function(object) validateEnum(object, ZZ)
# See RGCCTranslationUnit and RAutoGenRunTime....           
#           body(f)[[3]] = as.character(i@values)
           f = NULL
           def = setClass(name, contains = "character", validity = f, where = where)
           
         } else if(is(i, "ClassDefinition")) {

               #   def <- createTypeClass(i, types, where = where)                       

                 # make certain the types for the fields are defined
            repn = createClassRepresentation(i, types, namespaceDefs)
            m = sapply(repn, function(x)  is.null(getClassDef(x)) )
            if(any(m)) {
               if(verbose) 
                cat("Digressing to define", paste(repn[m], collapse = ", "), "for", name, "\n")

                # Now we recursively call this function to define these outstanding nodes.
                # We do have to worry about the depth of the call stack as this could grow
                # quite large if the classes are given in the wrong order.
                #XXX  only looking in first schema here.
              sapply(repn[m], function(x) {
                                   # look in all of the types 
                                 for(i in seq(along = types)) {
                                    type = types[[i]][[x]]
                                    if(!is.null(type))
	                                 break
                                 }

                                 if(!is.null(type))
                                    defClass(type)
                              })
            }

            if(is(i, "ExtendedClassDefinition")) {
                baseClass = mapSOAPTypeToS(i@base, types = types)
                if(is.null(getClassDef(baseClass)))
                   defClass(types[[1]][[baseClass]])
            } else
                baseClass = "VirtualSOAPClass"
            
            def = setClass(name, representation = repn, where = where, contains = baseClass)

         } else if(is(i, "ArrayType")) {
            def = createArrayClass(i, types, where = where)
         } else
            def = NULL

         if(!is.null(def))
           assign(name, def, classes)
     }    
  
      # for each schema, define each type.
     lapply(types,  function(schema)  lapply(schema, defClass))


     classes
}  

genSOAPClientInterface =
  #
  # def is from reading the WSDL file
  # name is the prefix for the methods
  # env is where to put the methods.
  #
function(operations = def@operations[[1]], def, name = def@name, env = new.env(),
           where = globalenv(), server = def@server,
            nameSpaces = def@nameSpaces, verbose = FALSE)  
{
 classes = NULL
 if(!is.null(where)) 
  classes = defineClasses(def@types, where, verbose)

 funs = list()
 for(i in operations) {
   if(verbose)
     cat("Operation", i@name, "\n")

   funs[[convertToSName(i@name)]] = createOperationDefinition(i, server, def@types, env, nameSpaces)
 }
 
 ans = new("SOAPClientInterface", functions = funs, classes = as.list(classes))

 ans
}


getXSDSchemaURIs =
function() {
  sapply(.SOAPDefaultNameSpaces, "[", "xsd")
}  

coerceArgumentCode =
function(id, type)
{
  name = convertToSName(id)

  if(length(type@nsuri) && type@nsuri %in% getXSDSchemaURIs()) {
     which = match(type@name, sapply(SOAPTypes, "[[", "type"))
     ifelse(!is.na(which),
             paste("as.",  names(SOAPTypes)[which], "(", id, ")", sep = ""),
            id)
  } else {
     paste("as(", id, ", '", type@name, "')", sep = "")
  }
}  


createOperationDefinition =
function(.operation, .defaultServer, .types, env = globalenv(), nameSpaces = NA)
{
   # resolve the parameter types
 .operation@parameters = lapply(.operation@parameters, resolve, .types)

 .operation@returnValue = rval = resolve(.operation@returnValue, .types)

 if(!is.null(.operation@header))
   .operation@header$type = resolve(.operation@header$type, .types)

   # Strip the namespace prefix.
 returnClass <- ifelse(rval@name %in% gsub(".*:", "", names(.types)), rval@name, NA)
 fixes = character(2)
 if(!is.na(returnClass)) {
    fixes[1] = "  ans <-\n"
    fixes[2] = paste("\tas(ans, '",  returnClass, "')\n", sep = "")
 }
 
 
 if(.operation@action == "")
     # NULL is not used, so just NULL for placeholder.
   .operation@action[1] = .SOAPAction(.operation@action, .operation@name, NULL, .operation@namespace)

 dotArgs = ""
 if(length(.operation@parameters)) {
   values = sapply(names(.operation@parameters),  function(x) coerceArgumentCode(x, .operation@parameters[[x]]))
   dotArgs = paste(paste(paste("'", names(.operation@parameters), "'", sep=""),
                                       values, sep = " = "),
                   collapse=", ")
 }

 
 txt = paste("function(",
              paste(c(sapply(names(.operation@parameters), convertToSName),
                      "server = .defaultServer, .convert = .operation@returnValue, .opts = list(), ..."),
                    collapse=",\n\t "),
                    paste(",\n\t nameSpaces = ", ifelse(is.na(nameSpaces), nameSpaces, dQuote(nameSpaces))),   
             ")\n{\n",
              fixes[1],
             "\t .SOAP(server, .operation@name, ",
                  "\n\t\t",         dotArgs,
                   ifelse(length(.operation@parameters), ", ", ""),
                  "\n\t\t",         "action = .operation@action, ",
                  "\n\t\t",         "xmlns = .operation@namespace, ",
                  "\n\t\t",         ".types = .operation@parameters, ",
                  "\n\t\t",         ".convert = .convert,",
                  "\n\t\t",         ".header = .header,",
                  "\n\t\t",         ".opts = .opts ",
                  ifelse(.operation@use["input"] == "literal", ",\n\t\t .literal = TRUE", ""),
                  ", ...",
                  ", nameSpaces = nameSpaces",
                  ")\n",
                  fixes[2],
                  "\n}", sep="")  
 f = eval(parse(text = txt)) 

 environment(f) <- e <- new.env()
 e$.operation <- .operation
 e$.header <- getSOAPRequestHeader(.operation@action, .server = .defaultServer)
 if(!is.null(.defaultServer))
    e$.defaultServer <- .defaultServer
 
 parent.env(e) = env
 
 attr(f, "source") <- txt
 
 rm(txt)
 rm(dotArgs)
 
 f
}


setGeneric("convertFromSOAP", function(val, type) standardGeneric("convertFromSOAP"))

setMethod("convertFromSOAP", "SOAPResult", 
          function(val, type) {
            node <- parseSOAP(val@content, asText = TRUE)
            node <- getReturnNode(node)
            
            fromSOAP(node, type = type)
          })



if(FALSE) {

  createTypeClass =
  # probably not called anymore as lifted into the defineClasses() function above.
    function(type, types, where = globalenv(), parentClass = "VirtualSOAPClass",
         namespaceDefs = list())
      {
        if(verbose)
          cat("[createTypeClass]", type$name, "\n")
        repn = createClassRepresentation(type$definition, types, namespaceDefs)
        setClass(type$name, representation = repn, where = where, contains = parentClass)
      }
}



createArrayClass =
  # Should be merged into createTypeClass.sl
function(type, types, where = globalenv(), parentClass = "VirtualSOAPClass")
{
#  name = type$definition@elementType
  if(is(type, "SOAPType")) {
    el = type@elType
    name = type@name
  } else {
    el = type$definition@elType
    name = type$name    
  }

  builtinClass = "list"

    # Want to see if this is a basic type in R, e.g. integer
  if(el@nsuri %in% getXSDSchemaURIs()) {
    #builtinClass = getArrayElementTypeFromName(name)
    which = match(name, sapply(SOAPTypes, "[[", "type"))
    if(!is.na(which))
      builtinClass = names(SOAPTypes)[which]
     #    if(is.na(builtinClass))
     # stop("can't identify array element type for SOAP definition")
  }
  
        #XXX want a typed list where the elements are checked. See DCOM code.
  setClass(name, contains = builtinClass, where = where)      

}

getRTypeFromSOAP =
function(el)
{
 i = match(paste("xsd:", el, sep = ""), sapply(SOAPTypes, function(x) x[["xsi:type"]]))
 if(!is.na(i))
   el = names(SOAPTypes)[i]
 
 el
}

getArrayElementTypeFromName =
function(name, stripArray = TRUE, convertToRType = TRUE)
{
 els = strsplit(name, ":")[[1]]
 if(length(els) > 1) {
   els = els[2]
 }
 if(stripArray)
   els = gsub("\\[\\]$", "", els)

 if(!convertToRType)
   return(els)

 getRTypeFromSOAP(els)
}  


newSOAPClass =
  #
  # Creates a new instance of the specified class (className) and populates its
  # fields with the values from the XML node.
  #
  #XXX converters is not used here yet.
function(node, className, converters = SOAPPrimitiveConverters, type = NULL)
{
 obj = new(className)
 
 for(i in slotNames(className)) {
   tmp = node[[i]]
   if(!is.null(type))
     slotType = type@slotTypes[[i]]
   else
     slotType = NULL

   slot(obj, i) <- fromSOAP(tmp, type = slotType, findReturn = FALSE) #XXX Need SOAP type here!
 }

 
  obj
}


createClassRepresentation =
 #
 # 
function(type, types, namespaceDefs = list())
{
   repn = lapply(type@slotTypes, mapSOAPTypeToS, types, namespaceDefs)
   nas = sapply(repn, is.na)
   if(any(nas))
     stop("problem resolving SOAP type ", names(type)[nas])

   repn
}

mapSOAPTypeToS =
  #
  # Take a SOAP type (by name or as a SOAPType object)
  # and find the name of the R data type that it maps to 
  # so that data type can be used in, e.g., a class representation
  # e.g. "int" goes to "integer".
  #
  #XXX Need to deal with namespaces. Tricky because they may have been
  # defined elsehwere, not in the top of the document. Need to match
  # the prefix to the actual URI and then compare URIs.
function(type, types = list(), namespaceDefs = list())
{
    if(is(type, "AttributeDef")) {
      # recover()
      warning("Not processing AttributeDef in mapSOAPTypeToS at present")
      return("character")
    }
  
    if(is(type, "SOAPType"))
      type = type@name

    if(length(type) == 1) { # just one entry
      type =  strsplit(type, ":")[[1]]
      if(length(type) == 2)  # so we have a namespace as well as the type name
               # call this function again with these two elements.
        return(mapSOAPTypeToS(type, types, namespaceDefs))
    }

    if(length(type) == 2) {
      ns = type[1]
      type = type[2]
    }

    if(length(grep("^ArrayOf", type)) > 0 && type %in% unlist(c(sapply(types, names)))) {
        # just give back the name of the R type (to be)
      return(type)
      # return(mapSOAPTypeToS(gsub("^ArrayOf", "", type), types, namespaceDefs))
    }


    which = sapply(SOAPTypes, function(x) x[["type"]] == type) #is this now correct?
    if(any(which))
      return((names(SOAPTypes)[which])[1])

#    if(length(grep(":", type)))
#      type = gsub("^.*:", "", type)
    
    if(length(types)) {

      if(is(types, "SchemaCollection"))
        for(i in types) {
           which = match(type, sapply(names(i), discardNamespace))
           if(!is.na(which))
               return(discardNamespace(names(i)[which]))
        }
    }

    stop("Can't match SOAP type ", type)
}



if(FALSE) {

# f = function(){}
# formals(f) <- alist(server=, kid=, threshold=, orgs=)
  
 body(f) <- substitute({
  val = .SOAP(server,
              .opName,
              action = .action,
              xmlns = .namespace,
              .types = ..types)
  }, 
  list(.opName= operation@name,
       .action = operation@action,
       .namespace = operation@namespace,
       ..types = operation@parameters))

  # now put the arguments in.
 e = body(f)[[2]]
 kk = e[4:length(e)]
# for(i in )
 k = body(f)[[2]][[3]]
}

# body(f) = substitute(body(f), list(.opname=operation@name))
